#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import asyncio
import os
import re
import unittest
from collections.abc import Iterator
from contextlib import contextmanager
from datetime import datetime, timezone
from io import BytesIO
from typing import Any
from urllib.parse import parse_qs, urlencode, urlparse

from ..data import HTTPHeaderDict, RequestMethod
from ..io.chunked_io_manager import DEFAULT_CHUNK_SIZE
from .http import AbstractTestRequestHandler, MockResponse, TestTransport
from .storage import TestWithStorage


class UrlGenerator:
    def __init__(self, base_url: str = "https://unittest.localhost/") -> None:
        self._base_url = base_url.removesuffix("/")
        self._n = 0
        self._current_url: str | None = None

    @property
    def n_calls(self) -> int:
        return self._n

    @property
    def current_url(self) -> str:
        assert self._current_url is not None
        return self._current_url

    async def get_new_url(self) -> str:
        self._n += 1
        self._current_url = (
            urlparse(self._base_url)
            ._replace(
                path="file.txt",
                query=urlencode(
                    [
                        ("attempt", self._n),
                        ("oct", self._n),
                        ("dt", datetime.now(timezone.utc).isoformat()),
                        ("blank", ""),
                    ]
                ),
            )
            .geturl()
        )
        return self.current_url


RANGE_HEADER_RE = re.compile(r"bytes=(?P<first>\d+)-(?P<last>\d+)")


class DownloadRequestHandler(AbstractTestRequestHandler):
    def __init__(self, data: bytes) -> None:
        self.data = data
        self._mutex = asyncio.Lock()
        self._buffer = BytesIO(data)
        self.size = len(data)

    async def handle_get(self, headers: HTTPHeaderDict) -> MockResponse:
        """Handle GET range requests for the test data.

        If the request is not a range request with a single range, return a 400 response.

        :param headers: The request headers.

        :return: A response containing the requested data.
        """
        requested = headers.get("Range", "")
        if match := RANGE_HEADER_RE.fullmatch(requested):
            requested_first = int(match.group("first"))
            requested_last = int(match.group("last"))
        else:
            return self.bad_request()

        requested_size = (requested_last - requested_first) + 1
        async with self._mutex:
            actual_first = self._buffer.seek(requested_first, os.SEEK_SET)
            data = self._buffer.read(requested_size)
            actual_last = self._buffer.tell() - 1
        response_headers = {
            "Content-Range": f"bytes {actual_first}-{actual_last}/{self.size}",
            "Content-Length": str(len(data)),
        }
        return MockResponse(status_code=206, body=data, reason="Partial Content", headers=response_headers)

    async def handle_head(self) -> MockResponse:
        """Handle HEAD requests for the test data.

        :return: A response containing the size of the test data and indicating support for range requests.
        """
        return MockResponse(
            status_code=200,
            reason="OK",
            headers={
                "Accept-Ranges": "bytes",
                "Content-Length": str(self.size),
            },
        )

    async def request(
        self,
        method: RequestMethod,
        url: str,
        headers: HTTPHeaderDict | None = None,
        post_params: list[tuple[str, str | bytes]] | None = None,
        body: object | str | bytes | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> MockResponse:
        match method:
            case RequestMethod.HEAD:
                return await self.handle_head()
            case RequestMethod.GET:
                return await self.handle_get(headers or HTTPHeaderDict())
            case _:
                return self.not_found()


class TestWithDownloadHandler(unittest.IsolatedAsyncioTestCase, TestWithStorage):
    def setUp(self) -> None:
        super().setUp()
        if not hasattr(self, "transport"):
            # Allowing for multiple inheritance, where the transport is set in a different setUp method.
            self.transport = TestTransport()
        self.url_generator = UrlGenerator()

    def setup_download_handler(self, data: bytes) -> DownloadRequestHandler:
        self.handler = DownloadRequestHandler(data)
        self.transport.set_request_handler(self.handler)
        return self.handler

    def assert_head_request_made(self, url: str) -> None:
        """Assert that a HEAD request was made to the given URL.

        :param url: The URL the HEAD request should have been made to.
        """
        self.transport.assert_any_request_made(RequestMethod.HEAD, url)

    def assert_range_request_made(self, url: str, offset: int, length: int) -> None:
        """Assert that a range request was made with the given offset and length.

        :param url: The URL the range request should have been made to.
        :param offset: The expected offset of the range request.
        :param length: The expected length of the range request.
        """
        self.transport.assert_any_request_made(
            RequestMethod.GET,
            url,
            headers={
                "Range": f"bytes={offset}-{offset + length - 1}",
            },
        )

    def assert_download_requests(self, url: str, chunk_size: int = DEFAULT_CHUNK_SIZE) -> None:
        """Assert that the correct requests were made to download the file.

        :param url: The URL the file should have been downloaded from.
        :param chunk_size: The expected chunk size for the requests.
        """
        filesize = len(self.handler.data)
        self.assert_head_request_made(url)
        for offset in range(0, filesize, chunk_size):
            self.assert_range_request_made(url, offset, min(chunk_size, filesize - offset))


class StorageDestinationRequestHandler(AbstractTestRequestHandler):
    def __init__(self) -> None:
        self._mutex = asyncio.Lock()
        self._buffer: dict[str, bytes] = {}
        self._block_list: list[str] | None = None
        self._expired_query_params: dict[str, str] | None = None

    def is_expired(self, query_params: dict[str, Any]) -> bool:
        """Check if the URL is expired."""
        if self._expired_query_params is None:
            return False
        current_query_params = {field: query_params[field] for field in self._expired_query_params}
        return self._expired_query_params == current_query_params

    async def get_committed(self) -> bytes:
        """Get the committed data from the buffer.

        :return: The committed data.
        """
        async with self._mutex:
            assert self._block_list is not None
            return b"".join(self._buffer[block_id] for block_id in self._block_list)

    async def handle_put_block(self, block_id: str, headers: HTTPHeaderDict, body: bytes) -> MockResponse:
        """Mock Storage PUT BLOCK operation.

        :param block_id: The block id from the query parameters.
        :param headers: The request headers.
        :param body: The request body.

        :return: A response indicating success.
        """
        async with self._mutex:
            assert block_id not in self._buffer, f"Block {block_id} already exists."
            self._buffer[block_id] = body

        return MockResponse(status_code=201, reason="Created")

    async def handle_put_block_list(self, headers: HTTPHeaderDict, body: bytes) -> MockResponse:
        """Mock Storage PUT BLOCK LIST operation.

        :param headers: The request headers.
        :param body: The request body.

        :return: A response indicating success.
        """
        async with self._mutex:
            assert headers.get("Content-Type") == "text/plain; charset=UTF-8"
            document = body.decode("utf-8")
            assert document.startswith('<?xml version="1.0" encoding="utf-8"?>\n<BlockList>\n') and document.endswith(
                "\n</BlockList>"
            ), f"Invalid XML document\n{document}"
            assert self._block_list is None, "Block list already committed."

            self._block_list = []
            for entry in document.split("\n")[2:-1]:
                assert entry.startswith("  <Latest>") and entry.endswith("</Latest>"), f"Invalid XML row: '{entry}'"
                block_id = entry[10:-9]
                assert block_id in self._buffer, f"Missing block with ID: '{block_id}'"
                self._block_list.append(block_id)

            all_block_ids = set(self._buffer.keys())
            committed_block_ids = set(self._block_list)
            assert all_block_ids == committed_block_ids, f"Missing blocks: {all_block_ids - committed_block_ids}"

        return MockResponse(status_code=201, reason="Created")

    async def handle_put(self, url: str, headers: HTTPHeaderDict, body: bytes) -> MockResponse:
        assert body is not None, "Body is required for PUT requests."
        assert "Content-Length" in headers, "Content-Length header is required for PUT requests."
        assert int(headers.get("Content-Length")) == len(body), "Content-Length header must match body length."
        uri = urlparse(url)
        query = parse_qs(uri.query)

        if self.is_expired(query):  # Mock expired URL.
            return MockResponse(status_code=403, reason="Forbidden")

        match query.get("comp", [""]):
            case ["block"]:
                try:
                    block_id = query["blockid"][0]
                except (KeyError, IndexError):
                    return self.bad_request()
                return await self.handle_put_block(block_id, headers, body)
            case ["blocklist"]:
                return await self.handle_put_block_list(headers, body)
            case _:
                return self.bad_request()

    async def request(
        self,
        method: RequestMethod,
        url: str,
        headers: HTTPHeaderDict | None = None,
        post_params: list[tuple[str, str | bytes]] | None = None,
        body: object | str | bytes | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> MockResponse:
        match method:
            case RequestMethod.PUT:
                return await self.handle_put(url, headers, body)
            case _:
                return self.not_found()

    @contextmanager
    def expired_url(self, url: str) -> Iterator[None]:
        """Context manager that simulates an expired URL."""
        uri = urlparse(url)
        self._expired_query_params = parse_qs(uri.query)
        yield
        self._expired_query_params = None


class TestWithUploadHandler(unittest.IsolatedAsyncioTestCase, TestWithStorage):
    def setUp(self) -> None:
        super().setUp()
        if not hasattr(self, "transport"):
            # Allowing for multiple inheritance, where the transport is set in a different setUp method.
            self.transport = TestTransport()
        self.handler = StorageDestinationRequestHandler()
        self.transport.set_request_handler(self.handler)
        self.url_generator = UrlGenerator()



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations

import asyncio
import contextlib
import logging
from abc import ABC, abstractmethod

from ..exceptions import RetryError

__all__ = [
    "BackoffExponential",
    "BackoffIncremental",
    "BackoffLinear",
    "BackoffMethod",
    "Retry",
    "RetryHandler",
]


class BackoffMethod(ABC):
    """Abstract base class for backoff methods.

    The only state is the configuration for the backoff method. One instance may be used for multiple requests.
    """

    def __init__(self, backoff_factor: int | float, max_delay: int | float = -1) -> None:
        """
        :param backoff_factor: The backoff factor to apply. This informs the size of the delay between attempts.
        :param max_delay: The maximum delay to apply. If the calculated delay is greater than this value, the maximum
            delay is used instead. If this value is negative, no maximum delay is applied.
        """
        self._backoff_factor = backoff_factor
        self._max_delay = max_delay

    @abstractmethod
    def _calculate_backoff_time(self, attempt_number: int) -> int | float:
        """Calculate the backoff time as a function of the number of attempts.

        :param attempt_number: The number of attempts that have been made.

        returns: Time to backoff for the current number of attempts
        """
        ...

    def get_backoff_time(self, attempt_number: int) -> int | float:
        """Get the backoff time as a function of the number of attempts.

        :param attempt_number: The number of attempts that have been made.

        returns: Time to backoff for the current number of attempts, or the maximum delay if the calculated delay is
            greater.
        """
        t = self._calculate_backoff_time(attempt_number)
        return t if self._max_delay < 0 else min(t, self._max_delay)


class BackoffLinear(BackoffMethod):
    """Linear retry delay."""

    def _calculate_backoff_time(self, attempt_number: int) -> int | float:
        return self._backoff_factor


class BackoffIncremental(BackoffMethod):
    """Incremental retry delay."""

    def _calculate_backoff_time(self, attempt_number: int) -> int | float:
        return self._backoff_factor * attempt_number


class BackoffExponential(BackoffMethod):
    """Exponential retry delay."""

    def _calculate_backoff_time(self, attempt_number: int) -> int | float:
        return self._backoff_factor * (2**attempt_number)


class _RetryIterator:
    def __init__(self, logger: logging.Logger, max_attempts: int, backoff_method: BackoffMethod) -> None:
        self.__logger = logger
        self.__max_attempts = max_attempts
        self.__backoff_method = backoff_method
        self.__current_handler: RetryHandler | None = None
        self.__errors = []

    def add_error(self, error: Exception) -> None:
        self.__logger.error(error)
        self.__logger.debug(f"{self.__current_handler} failed")
        self.__errors.append(error)
        if self.__current_handler._number >= self.__max_attempts:
            raise RetryError("Retry failed", self.__errors)

    async def __anext__(self) -> RetryHandler:
        if self.__current_handler is None:
            # Use a 1-based sequence for readability.
            next_handler = RetryHandler(self, 1)
        elif self.__current_handler.succeeded:
            raise StopAsyncIteration
        elif self.__current_handler._number >= self.__max_attempts:
            raise RetryError("Retry failed", self.__errors)
        else:
            # Calculate delay based on previous attempt.
            delay = self.__backoff_method.get_backoff_time(self.__current_handler._number)
            self.__logger.debug(f"Waiting {delay}s")
            await asyncio.sleep(delay)
            next_handler = self.__current_handler._next()

        self.__current_handler = next_handler

        return next_handler


class RetryHandler:
    """RetryHandler for retry attempts.

    A RetryHandler is created by a Retry object and is used to handle a single retry attempt. This handler should be
    used to suppress errors raised during the retry attempt, by using the suppress_errors context manager. If no errors
    are suppressed, the retry attempt is considered to have succeeded.
    """

    def __init__(self, iterator: _RetryIterator, attempt_number: int) -> None:
        """
        :param iterator: The iterator that created this handler.
        :param attempt_number: The number of preceding attempts.
        """
        self.__iterator = iterator
        self.__number = attempt_number
        self.__exception: Exception | None = None

    @contextlib.contextmanager
    def suppress_errors(self, excs: type[BaseException] | tuple[type[BaseException, ...]] | None = None) -> None:
        """Context manager to suppress errors raised during a retry attempt.

        :param excs: The exception types to suppress, defaults to all exceptions.
        """
        try:
            yield
        except Exception as exc:
            if excs is None or isinstance(exc, excs):
                self.__exception = exc
                self.__iterator.add_error(exc)
            else:
                raise

    @property
    def exception(self) -> Exception | None:
        """The exception raised during the retry attempt, if any."""
        return self.__exception

    def reset_counter(self) -> None:
        """Reset the attempt counter."""
        self.__number = 1

    @property
    def succeeded(self) -> bool:
        """Whether the retry attempt succeeded, determined by whether an exception was suppressed."""
        return self.__exception is None

    @property
    def failed(self) -> bool:
        """Whether the retry attempt failed."""
        return self.__exception is not None

    @property
    def _number(self) -> int:
        return self.__number

    def _next(self) -> RetryHandler:
        return RetryHandler(self.__iterator, self.__number + 1)

    def __str__(self) -> str:
        return f"Retry attempt #{self.__number}"


class Retry:
    """AsyncIterable retry implementation for handling exceptions in retryable logic.

    A single Retry object can be used to handle multiple requests. The only state in this class is the configuration.

    Usage::
        retry = Retry(logger=logging.getLogger(__name__), max_attempts=3, backoff_method=BackoffLinear(1))
        async for handler in retry:
            # do some things
            ...
            with handler.suppress_errors():
                # make request
                ...
            if handler.failed:
                # do some cleanup
                ...
    """

    def __init__(
        self,
        logger: logging.Logger,
        max_attempts: int = 3,
        backoff_method: BackoffMethod = BackoffExponential(backoff_factor=2),
    ) -> None:
        """Initialise a Retry object used when retrying after failures.

        :param logger: Logger instance for logging retry attempts.
        :param max_attempts: Maximum number of times to retry.
        :param backoff_method: Backoff method to apply.
        """
        if max_attempts < 1:
            raise ValueError("max_attempts must be greater than 0")
        self.__logger = logger
        self.__max_attempts = max_attempts
        self.__backoff_method = backoff_method

    def __aiter__(self) -> _RetryIterator:
        return _RetryIterator(self.__logger, self.__max_attempts, self.__backoff_method)



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations

from .authorizer import (
    AccessTokenAuthorizer,
    AuthorizationCodeAuthorizer,
    ClientCredentialsAuthorizer,
)
from .connector import OAuthConnector
from .data import AccessToken, AnyScopes, EvoScopes, OAuthScopes, Scopes
from .exceptions import OAuthError
from .oauth_redirect_handler import OAuthRedirectHandler

__all__ = [
    "AccessToken",
    "AccessTokenAuthorizer",
    "AnyScopes",
    "AuthorizationCodeAuthorizer",
    "ClientCredentialsAuthorizer",
    "EvoScopes",
    "OAuthConnector",
    "OAuthError",
    "OAuthRedirectHandler",
    "OAuthScopes",
    "Scopes",
]



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from uuid import UUID

from evo.discovery import Hub, Organization

BASE_URL = "http://unittest.localhost/"
ACCESS_TOKEN = "<not-a-real-token>"

HUB = Hub(
    code="hub-code",
    display_name="Hub Name",
    url=BASE_URL,
    services=["service1", "service2"],
)
ORG = Organization(id=UUID(int=1234), display_name="Organization Name", hubs=(HUB,))
WORKSPACE_ID = UUID(int=5678)



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Literal, TypeAlias
from uuid import UUID

from pydantic import ValidationError
from pydantic.type_adapter import TypeAdapter

from evo.common import APIConnector, HealthCheckType, Page, ServiceHealth, ServiceUser
from evo.common.utils import get_service_health, parse_order_by

from .data import (
    BasicWorkspace,
    BoundingBox,
    Coordinate,
    OrderByOperatorEnum,
    User,
    UserRole,
    Workspace,
    WorkspaceOrderByEnum,
    WorkspaceRole,
)
from .endpoints.api import AdminApi, GeneralApi, ThumbnailsApi, WorkspacesApi
from .endpoints.models import (
    BasicWorkspaceResponse,
    CreateWorkspaceRequest,
    GeometryTypeEnum,
    Label,
    RoleEnum,
    UpdateWorkspaceRequest,
    WorkspaceRoleOptionalResponse,
    WorkspaceRoleRequiredResponse,
)
from .endpoints.models import BoundingBox as PydanticBoundingBox
from .endpoints.models import Coordinate as PydanticCoordinate
from .endpoints.models import User as PydanticUser
from .endpoints.models import UserRole as PydanticUserRole

WorkspaceOrderByLiteral: TypeAlias = Literal["name", "created_at", "updated_at", "user_role"]
OrderByOperatorLiteral: TypeAlias = Literal["asc", "desc"]


class WorkspaceAPIClient:
    """
    The Workspace Service API client.
    """

    def __init__(self, connector: APIConnector, org_id: UUID) -> None:
        self._connector = connector
        self._org_id = org_id
        self._workspaces_api = WorkspacesApi(connector)
        self._admin_api = AdminApi(connector)
        self._general_api = GeneralApi(connector)
        self._thumbnails_api = ThumbnailsApi(connector)

    async def get_service_health(self, check_type: HealthCheckType = HealthCheckType.FULL) -> ServiceHealth:
        """Get the health of the workspace service.
        :param check_type: The type of health check to perform.
        :return: A ServiceHealth object.
        :raises EvoAPIException: If the API returns an unexpected status code.
        :raises ClientValueError: If the response is not a valid service health check response.
        """
        return await get_service_health(self._connector, "workspace", check_type=check_type)

    @staticmethod
    def __parse_bounding_box(model: PydanticBoundingBox) -> BoundingBox:
        def convert_coordinate(pydantic_coordinate: PydanticCoordinate) -> Coordinate:
            return Coordinate(latitude=pydantic_coordinate.root[1], longitude=pydantic_coordinate.root[0])

        coordinates = [[convert_coordinate(coord) for coord in coord_list] for coord_list in model.coordinates]

        return BoundingBox(coordinates=coordinates, type=str(model.type.value))

    def __parse_workspace_model(
        self, model: WorkspaceRoleOptionalResponse | WorkspaceRoleRequiredResponse
    ) -> Workspace:
        bounding_box = None
        if model.bounding_box:
            bounding_box = self.__parse_bounding_box(model.bounding_box)

        return Workspace(
            id=model.id,
            org_id=self._org_id,
            hub_url=self._connector.base_url,
            display_name=model.name,
            description=model.description,
            user_role=WorkspaceRole[str(model.current_user_role.value)] if model.current_user_role else None,
            created_at=model.created_at,
            created_by=ServiceUser.from_model(model.created_by),
            updated_at=model.updated_at,
            updated_by=ServiceUser.from_model(model.updated_by),
            bounding_box=bounding_box,
            default_coordinate_system=model.default_coordinate_system,
            labels=model.labels,
        )

    @staticmethod
    def __parse_workspace_basic_model(model: BasicWorkspaceResponse) -> BasicWorkspace:
        return BasicWorkspace(
            id=model.id,
            display_name=model.name,
        )

    @staticmethod
    def __parse_user_role_model(model: PydanticUserRole) -> UserRole:
        return UserRole(user_id=model.user_id, role=WorkspaceRole[str(model.role.value)])

    @staticmethod
    def __parse_user_model(model: PydanticUser) -> User:
        return User(
            user_id=model.user_id,
            role=WorkspaceRole[str(model.role.value)],
            email=model.email,
            full_name=model.full_name,
        )

    async def list_user_roles(
        self,
        workspace_id: UUID,
        filter_user_id: UUID | None = None,
    ) -> list[User]:
        response = await self._workspaces_api.list_user_roles(
            org_id=str(self._org_id),
            workspace_id=str(workspace_id),
            filter_user_id=str(filter_user_id) if filter_user_id else None,
        )

        return [self.__parse_user_model(item) for item in response.results]

    async def get_current_user_role(
        self,
        workspace_id: UUID,
    ) -> UserRole:
        response = await self._workspaces_api.get_current_user_role(
            org_id=str(self._org_id),
            workspace_id=str(workspace_id),
        )

        return self.__parse_user_role_model(response)

    async def assign_user_role(
        self,
        workspace_id: UUID,
        user_id: UUID,
        role: WorkspaceRole,
    ) -> UserRole:
        assign_role_request = PydanticUserRole(user_id=user_id, role=RoleEnum[str(role.name)])
        response = await self._workspaces_api.assign_user_role(
            org_id=str(self._org_id),
            workspace_id=str(workspace_id),
            assign_role_request=assign_role_request,
        )

        return self.__parse_user_role_model(response)

    async def delete_user_role(
        self,
        workspace_id: UUID,
        user_id: UUID,
    ) -> None:
        await self._workspaces_api.delete_user_role(
            org_id=str(self._org_id),
            workspace_id=str(workspace_id),
            user_id=str(user_id),
        )

    async def list_workspaces(
        self,
        limit: int | None = None,
        offset: int | None = None,
        order_by: dict[WorkspaceOrderByEnum, OrderByOperatorEnum]
        | dict[WorkspaceOrderByLiteral, OrderByOperatorLiteral]
        | None = None,
        filter_created_by: UUID | None = None,
        created_at: str | None = None,
        updated_at: str | None = None,
        name: str | None = None,
        deleted: bool | None = None,
        filter_user_id: UUID | None = None,
    ) -> Page[Workspace]:
        parsed_order_by = parse_order_by(order_by)  # type: ignore
        if not offset:
            offset = 0
        response = await self._workspaces_api.list_workspaces(
            org_id=str(self._org_id),
            limit=limit,
            offset=offset,
            order_by=parsed_order_by,
            filter_created_by=str(filter_created_by) if filter_created_by else None,
            created_at=created_at,
            updated_at=updated_at,
            name=name,
            deleted=deleted,
            filter_user_id=str(filter_user_id) if filter_user_id else None,
        )

        return Page(
            offset=offset,
            limit=limit,
            total=response.links.total,
            items=[self.__parse_workspace_model(item) for item in response.results],
        )

    async def list_all_workspaces(
        self,
        limit: int | None = None,
        offset: int | None = None,
        order_by: dict[WorkspaceOrderByEnum, OrderByOperatorEnum] | None = None,
        filter_created_by: UUID | None = None,
        created_at: str | None = None,
        updated_at: str | None = None,
        name: str | None = None,
        deleted: bool | None = None,
        filter_user_id: UUID | None = None,
    ) -> list[Workspace]:
        workspaces: list[Workspace] = []
        if offset is None:
            offset = 0
        if limit is None:
            limit = 50

        while True:
            workspace_page = await self.list_workspaces(
                limit=limit,
                offset=offset,
                order_by=order_by,
                filter_created_by=filter_created_by,
                created_at=created_at,
                updated_at=updated_at,
                name=name,
                deleted=deleted,
                filter_user_id=filter_user_id,
            )
            workspaces += workspace_page.items()
            offset += limit
            if offset >= workspace_page.total:
                break

        return sorted(workspaces, key=lambda x: x.display_name)

    async def list_workspaces_summary(
        self,
        limit: int | None = None,
        offset: int | None = None,
        order_by: dict[WorkspaceOrderByEnum, OrderByOperatorEnum]
        | dict[WorkspaceOrderByLiteral, OrderByOperatorLiteral]
        | None = None,
        filter_created_by: UUID | None = None,
        created_at: str | None = None,
        updated_at: str | None = None,
        name: str | None = None,
        deleted: bool | None = None,
        filter_user_id: UUID | None = None,
    ) -> Page[BasicWorkspace]:
        """
        Get an optionally paginated list of basic workspaces with optional filtering and sorting.

        This method provides faster performance than list_workspaces() or list_all_workspaces()
        by returning BasicWorkspace objects with minimal data instead of full Workspace objects.
        """
        parsed_order_by = parse_order_by(order_by)

        response = await self._workspaces_api.list_workspaces_summary(
            org_id=str(self._org_id),
            limit=limit,
            offset=offset,
            order_by=parsed_order_by,
            filter_created_by=str(filter_created_by) if filter_created_by else None,
            created_at=created_at,
            updated_at=updated_at,
            name=name,
            deleted=deleted,
            filter_user_id=str(filter_user_id) if filter_user_id else None,
        )

        return Page(
            offset=offset or 0,
            limit=limit or response.links.total,
            total=response.links.total,
            items=[self.__parse_workspace_basic_model(item) for item in response.results],
        )

    async def get_workspace(self, workspace_id: UUID, deleted: bool = False) -> Workspace:
        response = await self._workspaces_api.get_workspace(
            org_id=str(self._org_id), workspace_id=str(workspace_id), deleted=deleted
        )
        return self.__parse_workspace_model(response)

    async def delete_workspace(
        self,
        workspace_id: UUID,
    ) -> None:
        """Delete workspaces by workspace id.

        :param workspace_id: The workspace id to delete.

        :returns: An empty response.
        """
        await self._workspaces_api.delete_workspace(
            org_id=str(self._org_id),
            workspace_id=str(workspace_id),
        )

    async def create_workspace(
        self,
        name: str,
        bounding_box_coordinates: list[tuple[float, float]] | None = None,
        default_coordinate_system: str | None = None,
        description: str | None = None,
        labels: list[str] | None = None,
    ) -> Workspace:
        """Create a new workspace.

        :param name: The name of the workspace.
        :param bounding_box_coordinates: The coordinates list corresponding to the bounding box of the workspace.
            The coordinates list should be in the format of [[longitude, latitude], [longitude, latitude], ...]
        :param default_coordinate_system: The default coordinate system of the workspace.
        :param description: The description of the workspace.
        :param labels: The labels of the workspace.

        :returns: The created workspace response.
        """
        # apply validation on the values
        if description is None:
            description = ""
        if default_coordinate_system is None:
            default_coordinate_system = ""
        if labels is not None and len(labels) > 20:
            raise ValueError("The labels field must contain 20 or fewer items.")

        if bounding_box_coordinates is not None:
            try:
                bounding_box = PydanticBoundingBox(
                    type=GeometryTypeEnum.Polygon,
                    coordinates=[bounding_box_coordinates],
                )
            except ValidationError as e:
                raise ValueError(
                    "Invalid bounding box coordinates! Ensure that the bounding box coordinate is in the format of [[longitude, latitude], [longitude, latitude], ...]"
                ) from e
        else:
            bounding_box = None

        create_workspace_request = CreateWorkspaceRequest(
            name=name,
            bounding_box=bounding_box,
            default_coordinate_system=default_coordinate_system,
            description=description,
            labels=labels,
        )
        model = await self._workspaces_api.create_workspace(
            org_id=str(self._org_id), create_workspace_request=create_workspace_request
        )
        return self.__parse_workspace_model(model)

    async def update_workspace(
        self,
        workspace_id: UUID,
        name: str | None = None,
        bounding_box_coordinates: list[tuple[float, float]] | None = None,
        default_coordinate_system: str | None = None,
        description: str | None = None,
        labels: list[str] | None = None,
    ) -> Workspace:
        """Update an existing workspace.

        :param workspace_id: The workspace id to update.
        :param name: The name of the workspace.
        :param bounding_box_coordinates: The coordinates list corresponding to the bounding box of the workspace.
            The coordinates list should be in the format of [[longitude, latitude], [longitude, latitude], ...]
        :param default_coordinate_system: The default coordinate system of the workspace.
        :param description: The description of the workspace.
        :param labels: The labels of the workspace.

        :returns: The updated workspace response.
        """
        # apply validation on the values
        update_workspace_request = UpdateWorkspaceRequest()
        if description is not None:
            update_workspace_request.description = description
        if default_coordinate_system is not None:
            update_workspace_request.default_coordinate_system = default_coordinate_system
        if name is not None:
            update_workspace_request.name = name
        if labels is not None:
            if len(labels) > 20:
                raise ValueError("The labels field must contain 20 or fewer items.")
            labels_adapter = TypeAdapter(list[Label])
            update_workspace_request.labels = labels_adapter.validate_python(labels)

        if bounding_box_coordinates is not None:
            try:
                bounding_box = PydanticBoundingBox(
                    type=GeometryTypeEnum.Polygon,
                    coordinates=[bounding_box_coordinates],
                )
                update_workspace_request.bounding_box = bounding_box
            except ValidationError as e:
                raise ValueError(
                    "Invalid bounding box coordinates! Ensure that the bounding box coordinate is in the format of [[longitude, latitude], [longitude, latitude], ...]"
                ) from e
        model = await self._workspaces_api.update_workspace(
            org_id=str(self._org_id), workspace_id=str(workspace_id), update_workspace_request=update_workspace_request
        )
        return self.__parse_workspace_model(model)



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from pure_interface import Interface

__all__ = [
    "IDestination",
    "ISource",
]


class IDestination(Interface):
    """A local or remote destination for managed file IO.

    IDestination implementations should raise a subtype of ChunkedIOError when a recoverable error occurs. The specific
    error will depend on the implementation and must be capable of recovering the IDestination from the failure state.
    """

    async def write_chunk(self, offset: int, data: bytes) -> None:
        """Write raw data to the destination at the provided offset"""


class ISource(Interface):
    """A local or remote source for managed file IO.

    ISource implementations should raise a subtype of ChunkedIOError when a recoverable error occurs. The specific
    error will depend on the implementation and must be capable of recovering the ISource from the failure state.
    """

    async def get_size(self) -> int:
        """Get the size of the source data"""

    async def read_chunk(self, offset: int, length: int) -> bytes:
        """Read <length> bytes of raw data from the source, starting at the given offset"""



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from dataclasses import dataclass
from uuid import UUID

__all__ = [
    "Hub",
    "Organization",
]


@dataclass(frozen=True, kw_only=True)
class Hub:
    """Hub metadata."""

    url: str
    """Hub URL."""

    code: str
    """Hub shortcode."""

    display_name: str
    """Hub display name."""

    services: tuple[str, ...]
    """List of service codes."""


@dataclass(frozen=True, kw_only=True)
class Organization:
    """License holder organization metadata."""

    id: UUID
    """Organization ID."""

    display_name: str
    """Organization display name."""

    hubs: tuple[Hub, ...]



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import inspect
import shutil
import unittest
from pathlib import Path

from ..data import Environment
from ..utils import Cache
from .consts import BASE_URL, ORG, WORKSPACE_ID


class TestWithStorage(unittest.TestCase):
    CACHE_DIR: Path

    @classmethod
    def setUpClass(cls) -> None:
        cache_dir = Path(inspect.getfile(cls)).parent.resolve() / f".{cls.__name__.lower()}_cache"

        def _cleanup_cache() -> None:
            """Fail-safe cleanup of the cache directory."""
            shutil.rmtree(cache_dir, ignore_errors=True)

        cls.addClassCleanup(_cleanup_cache)
        cache_dir.mkdir(parents=True, exist_ok=True)

        # Write gitignore file in case cleanup fails.
        gitignore_file = cache_dir / ".gitignore"
        gitignore_file.write_text("*\n", encoding="utf-8")

        cls.CACHE_DIR = cache_dir

    def setUp(self) -> None:
        self.cache = Cache(self.CACHE_DIR)
        self.environment = Environment(hub_url=BASE_URL, org_id=ORG.id, workspace_id=WORKSPACE_ID)



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import shutil
import tempfile
from collections.abc import Iterator
from contextlib import contextmanager
from pathlib import Path
from typing import TypeAlias
from uuid import uuid5

from evo.common.data import Environment
from evo.common.exceptions import StorageFileExistsError, StorageFileNotFoundError
from evo.common.interfaces import ICache

__all__ = ["Cache"]


FileName: TypeAlias = str | Path


class Cache(ICache):
    """An optional utility to manage cache directories for API data.

    The cache may be used to store transient binary data, such as files from File API or parquet files from
    the Geoscience Object API.
    """

    def __init__(self, root: FileName, mkdir: bool = False) -> None:
        """
        :param root: The root cache directory.
        :param mkdir: If True, create the cache directory if it does not exist.
        """
        match root:
            case Path():
                pass  # No conversion necessary.
            case str():
                root = Path(root)
            case _:
                raise TypeError(f"Expected a Path or str, got {root!r}")

        root = root.resolve()

        if mkdir:
            root.mkdir(parents=True, exist_ok=True)
        elif not root.exists():
            raise StorageFileNotFoundError(f"'{root}' does not exist.")

        if not root.is_dir():
            raise StorageFileExistsError(f"'{root}' is not a directory.")

        self._root = root

    @property
    def root(self) -> Path:
        """The absolute path to the root cache directory."""
        return self._root

    def get_location(self, environment: Environment, scope: str) -> Path:
        """Get the cache location for the specified environment and scope.

        :param environment: The environment used to determine the cache location.
        :param scope: The scope used to determine the cache location.

        :returns: The absolute path to the cache location.

        :raises StorageFileExistsError: If the location already exists, and it is not a directory.
        """
        cache_id = uuid5(environment.workspace_id, scope)
        storage = self.root / str(cache_id)

        try:
            storage.mkdir()
        except FileExistsError as err:
            if not storage.is_dir():
                raise StorageFileExistsError(f"'{storage}' is not a directory.") from err

        return storage

    def clear_cache(self, environment: Environment | None = None, scope: str | None = None) -> None:
        """Clear the cache for the specified environment and scope.

        If the environment and the scope is None, clear the entire cache.

        :param environment: The environment of the cache location. If None, clear the entire cache.
        :param scope: The scope of the cache location. If None, clear the entire cache.

        :raises ValueError: If either environment or scope is None, but not both.
        """
        if scope is None and environment is None:
            # Leave the root directory in place.
            targets = list(self.root.iterdir())
        elif scope is None or environment is None:
            raise ValueError("environment and scope must be specified together.")
        else:
            targets = [self.get_location(environment, scope)]

        for target in targets:
            if target.is_dir():
                shutil.rmtree(target)
            else:
                target.unlink()

    @contextmanager
    def temporary_location(self) -> Iterator[Path]:
        """Create a temporary cache directory.

        Uses tempfile.TemporaryDirectory internally to create a temporary directory within the cache root.
        The temporary directory is removed when the context manager exits.

        :returns: The absolute path to the temporary cache directory.
        """
        with tempfile.TemporaryDirectory(dir=self.root) as temp:
            yield Path(temp)



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations

from datetime import datetime, timedelta, timezone
from enum import Enum
from typing import TYPE_CHECKING, Literal, Optional, TypeAlias

from pydantic import BaseModel, Field
from typing_extensions import deprecated

__all__ = [
    "AccessToken",
    "AnyScopes",
    "EvoScopes",
    "OAuthScopes",
]

_ALLOWABLE_CLOCK_DRIFT = timedelta(minutes=5)


def _utcnow() -> datetime:
    """Get the current UTC time.

    :return: The current UTC time.
    """
    return datetime.now(timezone.utc)


class Scopes(str):
    """A space-separated list of OAuth scopes."""

    def __new__(cls, *members: str) -> Scopes:
        if len(members) == 0:
            raise ValueError("At least one scope must be provided")

        # Get unique members.
        unique_members = set()
        for member in members:
            # Each member may itself be a space-separated list of scopes.
            unique_members.update(member.strip().split(" "))

        inst = str.__new__(cls, " ".join(sorted(unique_members)))
        return inst

    @property
    def members(self) -> tuple[str, ...]:
        return tuple(sorted(self.split(" ")))

    def __repr__(self) -> str:
        return f"Scopes{self.members!r}"

    def __or__(self, other: str) -> Scopes:
        # Support Scopes | str
        if isinstance(other, str):
            return Scopes(self, other)
        return NotImplemented

    def __ror__(self, other: str) -> Scopes:
        # Support str | Scopes
        return self | other

    def __contains__(self, key: str) -> bool:
        members = set(self.members)
        others = set(Scopes(key).members)
        return others.issubset(members)


class EvoScopes(Scopes, Enum):
    """Public OAuth scopes for authenticating against Evo APIs.

    https://developer.seequent.com/docs/guides/getting-started/apps-and-tokens#about-evo-access-tokens
    """

    offline_access = Scopes("offline_access")

    # Evo Scopes
    evo_discovery = Scopes("evo.discovery")
    """Required for access to the Evo Discovery API, both discovery and token endpoints."""

    evo_workspace = Scopes("evo.workspace")
    """Required for access to the Workspaces API."""

    evo_audit = Scopes("evo.audit")
    """Required for access to the Audit API."""

    evo_blocksync = Scopes("evo.blocksync")
    """Required for access to the BlockSync API."""

    evo_object = Scopes("evo.object")
    """Required for access to the Geoscience Object API."""

    evo_file = Scopes("evo.file")

    # Useful combinations
    """Default scopes for Evo applications.

    Extra scopes to be added as needed include `evo_blocksync`, `evo_object`, or `evo_file`, etc.
    """
    default = evo_discovery | evo_workspace

    """All scopes for Evo applications."""
    all_evo = default | evo_blocksync | evo_object | evo_file

    def __str__(self) -> str:
        return self.value

    def __repr__(self) -> str:
        return Scopes.__repr__(self)


if TYPE_CHECKING:
    # Subclassing Enum is not supported at runtime, but it was the most elegant way to mark the old name as deprecated.
    # At runtime `OAuthScopes` is simply an alias for `EvoScopes`, and a deprecated warning will not be emitted.

    @deprecated("Use EvoScopes instead.")
    class OAuthScopes(EvoScopes): ...
else:
    OAuthScopes: TypeAlias = EvoScopes


AnyScopes: TypeAlias = str | Scopes | EvoScopes


class AccessToken(BaseModel):
    """An access token model, returned by the OAuth server.
    https://www.rfc-editor.org/rfc/rfc6749#section-5.1
    """

    token_type: Literal["Bearer"]

    access_token: str
    """The access token issued by the authorization server."""

    refresh_token: Optional[str] = None
    """The refresh token, which can be used to obtain new access tokens using the same authorization grant. Refresh tokens are only returned if you request the "offline access" scope."""

    expires_in: Optional[int] = None
    """The lifetime in seconds of the access token."""

    issued_at: datetime = Field(default_factory=_utcnow)
    """The time at which the token response was received."""

    @property
    def expires_at(self) -> datetime | None:
        """The time at which the token expires, or None if the token lifetime is unknown.

        A small margin of error may occur due to the time taken to process the token response.
        """
        if self.expires_in is None:
            return None
        return self.issued_at + timedelta(seconds=self.expires_in)

    @property
    def ttl(self) -> int | None:
        """The time-to-live (TTL) of this access token in seconds, or None if the token lifetime is unknown.

        If the token is expired, the TTL will be 0.
        """
        if self.expires_at is None:
            return None
        ttl = self.expires_at - _utcnow()
        return max(round(ttl.total_seconds()), 0)

    @property
    def is_expired(self) -> bool:
        """Whether this access token has expired.

        True if the access token has expired, False otherwise.
        If the token lifetime is unknown, this method will always return False.
        """
        if (expiry := self.expires_at) is not None:
            return _utcnow() > expiry
        else:
            return False



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations

from pathlib import Path
from types import TracebackType

from pure_interface import Interface

from .data import Environment, HTTPHeaderDict, HTTPResponse, RequestMethod

__all__ = [
    "IAuthorizer",
    "ICache",
    "IFeedback",
    "ITransport",
]


class ITransport(Interface):
    """Interface for HTTP Transport.

    ITransport is responsible for sending HTTP requests and receiving responses. The open and close methods are
    used to manage the connection state. The request method is used to send an HTTP request and receive a response.

    An internal counter should be incremented when open is called and decremented when close is called. The transport
    should only release resources when the counter reaches zero.
    """

    async def open(self) -> None:
        """Open the HTTP transport.

        This method should be called before sending any requests. ITransport implementations should be reentrant,
        meaning that calling open multiple times should not have any side effects. Resources that are consumed by the
        transport should be retained until the close method is called the same number of times as open.

        Reopening a transport object that has previously been closed should work the same way as opening a new transport
        object.
        """
        ...  # pragma: no cover

    async def close(self) -> None:
        """Close the transport.

        This method should be called when the HTTP transport is no longer needed. Because ITransport implementations
        should be reentrant, calling close does not guarantee that the transport is closed immediately. Instead,
        resources should be retained as long as there is an open call that has not been matched by a close call.
        """
        ...  # pragma: no cover

    async def __aenter__(self) -> ITransport:
        """ITransport should be an asynchronous context manager that opens the transport when entered, and closes it when
        exited.

        This allows the transport to be used in an async with statement.
        """
        ...

    async def __aexit__(
        self, exc_type: type[Exception] | None, exc_val: Exception | None, exc_tb: TracebackType | None
    ) -> None:
        """ITransport should be an asynchronous context manager that opens the transport when entered, and closes it when
        exited.

        This allows the transport to be used in an async with statement.
        """
        ...

    async def request(
        self,
        method: RequestMethod,
        url: str,
        headers: HTTPHeaderDict | None = None,
        post_params: list[tuple[str, str | bytes]] | None = None,
        body: object | str | bytes | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> HTTPResponse:
        """Send an asynchronous request.

        ITransport implementations *MUST NOT* automatically follow redirects. Redirects have special semantics and will
        be handled by individual clients.

        :param method: HTTP request method.
        :param url: HTTP request url.
        :param headers: Http request headers.
        :param body: Request body.
        :param post_params: Request post parameters, `application/x-www-form-urlencoded` and `multipart/form-data`.
        :param request_timeout: Timeout setting for this request. If one number provided, it will be total request
            timeout. It can also be a pair (tuple) of (connection, read) timeouts.

        :return: The HTTP response object.

        :raise ClientValueError: If `body` and `post_params` are both provided.
        :raise TransportError: If the underlying implementation encounters an error.
        """
        ...  # pragma: no cover


class IAuthorizer(Interface):
    """Interface for authorizing HTTP requests.

    IAuthorizer is responsible for providing the necessary headers to authorize HTTP requests. The get_default_headers
    method should return appropriate authorization headers for the target environment. The refresh_token method will be
    called when authorizing a request fails. Implementations should refresh the token and update the default_headers.

    Ideally, an async mutex should be used to ensure that get_default_headers does not return an expired token while
    refresh_token is running.
    """

    async def get_default_headers(self) -> HTTPHeaderDict:
        """Get the default headers for authorizing HTTP requests.

        Depending on the operating environment, different headers may be required to authorize requests. This interface
        allows the authorizer implementation to provide the necessary headers for authorization based on the environment
        the implementation is intended for.

        :return: A header dictionary to be included in the HTTP request by default.
        """
        ...  # pragma: no cover

    async def refresh_token(self) -> bool:
        """Refresh the authorization token.

        This method will be called when the authorization token has expired. Implementations should refresh the token
        and update the default_headers property with the new token.

        :return: True if the token was successfully refreshed, False otherwise.
        """
        ...  # pragma: no cover


class IFeedback(Interface):
    """An interface for providing live feedback to the user."""

    def progress(self, progress: float, message: str | None = None) -> None:
        """Progress the feedback and update the text to message.

        This can raise an exception to cancel the current operation.

        :param progress: A float between 0 and 1 representing the progress of the operation as a percentage.
        :param message: An optional message to display to the user.
        """
        ...  # pragma: no cover


class ICache(Interface):
    """An interface for managing cache directories for service data.

    The cache may be used to store transient binary data, such as files from File API or parquet files from
    the Geoscience Object API.
    """

    root: Path
    """The absolute path to the root cache directory."""

    def get_location(self, environment: Environment, scope: str) -> Path:
        """Get the cache location for the specified environment and scope.

        :param environment: The environment used to determine the cache location.
        :param scope: The scope used to determine the cache location.

        :returns: The absolute path to the cache location.

        :raises StorageFileExistsError: If the location already exists, and it is not a directory.
        """
        ...

    def clear_cache(self, environment: Environment | None = None, scope: str | None = None) -> None:
        """Clear the cache for the specified environment and scope.

        If the environment and the scope is None, clear the entire cache.

        :param environment: The environment of the cache location. If None, clear the entire cache.
        :param scope: The scope of the cache location. If None, clear the entire cache.

        :raises ValueError: If either environment or scope is None, but not both.
        """
        ...



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations

import asyncio
from typing import Any, TypeVar
from uuid import UUID

from evo import logging
from evo.common import APIConnector, BaseAPIClient, Environment
from evo.common.exceptions import SelectionError
from evo.common.interfaces import IAuthorizer, ITransport
from evo.discovery import DiscoveryAPIClient, Hub, Organization
from evo.workspaces import Workspace, WorkspaceAPIClient

__all__ = ["ServiceManager"]

logger = logging.getLogger("common.utils.manager")


# Generic type variable for the client factory method.
T_client = TypeVar("T_client", bound=BaseAPIClient)


class _State:
    """A simple state manager for managing the current selection of organizations, hubs, and workspaces."""

    def __init__(
        self,
        organizations: list[Organization],
        *,
        workspaces: list[Workspace] | None = None,
        selected_org_id: UUID | None = None,
        selected_hub_code: str | None = None,
        selected_workspace_id: UUID | None = None,
    ) -> None:
        """
        :param organizations: A list of organizations.
        :param workspaces: A list of workspaces.
        :param selected_org_id: The ID of the selected organization.
        :param selected_hub_code: The code of the selected hub.
        :param selected_workspace_id: The ID of the selected workspace.

        :raises SelectionError: If the selected organization, hub, or workspace is invalid.
        """
        self._organizations_by_id = {org.id: org for org in organizations}
        self._select_org(selected_org_id)
        self._select_hub(selected_hub_code)

        if workspaces is not None and not isinstance(self.organization, Organization):
            raise SelectionError("Cannot list workspaces without an organization.")

        if workspaces is not None and not isinstance(self.hub, Hub):
            raise SelectionError("Cannot list workspaces without a hub.")

        self._workspaces_by_id = {ws.id: ws for ws in workspaces} if workspaces is not None else {}
        self._select_workspace(selected_workspace_id)

    def _select_org(self, org_id: UUID | None) -> None:
        if org_id is not None and org_id not in self._organizations_by_id:
            raise SelectionError(f"Invalid organization ID: {org_id!r}")
        else:
            self._selected_org_id = org_id

        org = self.organization
        self._hubs_by_code = {hub.code: hub for hub in org.hubs} if isinstance(org, Organization) else {}

    def _select_hub(self, hub_code: str | None) -> None:
        if hub_code is not None and not isinstance(self.organization, Organization):
            raise SelectionError("Cannot select a hub without an organization.")
        elif hub_code is not None and hub_code not in self._hubs_by_code:
            raise SelectionError(f"Invalid hub code: {hub_code!r}")
        else:
            self._selected_hub_code = hub_code

    def _select_workspace(self, workspace_id: UUID | None) -> None:
        if workspace_id is not None and not isinstance(self.hub, Hub):
            raise SelectionError("Cannot select a workspace without a hub.")
        elif workspace_id is not None and workspace_id not in self._workspaces_by_id:
            raise SelectionError(f"Invalid workspace ID: {workspace_id!r}")
        else:
            self._selected_workspace_id = workspace_id

    @property
    def organizations(self) -> list[Organization]:
        """Get a list of organizations sorted by display name."""
        return sorted(self._organizations_by_id.values(), key=lambda org: org.display_name)

    @property
    def organization(self) -> Organization | None:
        """Get the currently selected organization."""
        return self._organizations_by_id[self._selected_org_id] if self._selected_org_id is not None else None

    @property
    def hubs(self) -> list[Hub]:
        """Get a list of hubs sorted by display name."""
        return sorted(self._hubs_by_code.values(), key=lambda hub: hub.display_name)

    @property
    def hub(self) -> Hub | None:
        """Get the currently selected hub."""
        return self._hubs_by_code[self._selected_hub_code] if self._selected_hub_code is not None else None

    @property
    def workspaces(self) -> list[Workspace]:
        """Get a list of workspaces sorted by display name."""
        return sorted(self._workspaces_by_id.values(), key=lambda ws: ws.display_name)

    @property
    def workspace(self) -> Workspace | None:
        """Get the currently selected workspace."""
        return self._workspaces_by_id[self._selected_workspace_id] if self._selected_workspace_id is not None else None

    def with_organization(self, org_id: UUID | None) -> _State:
        """Create a new state with the specified organization selected.

        :param org_id: The ID of the organization to select, or `None` to clear the current selection.

        :returns: The new state, or the current state if the organization was not changed.

        :raises SelectionError: If the organization is invalid.
        """
        logger.debug("Selecting organization")
        if org_id == getattr(self.organization, "id", None):
            logger.debug("Organization was not changed.")
            return self
        else:
            logger.debug("Organization was changed.")
            return _State(self.organizations, selected_org_id=org_id)

    def with_hub(self, hub_code: str | None) -> _State:
        """Create a new state with the specified hub selected.

        :param hub_code: The code of the hub to select, or `None` to clear the current selection.

        :returns: The new state, or the current state if the hub was not changed.

        :raises SelectionError: If the hub is invalid.
        """
        logger.debug("Selecting hub")
        if hub_code == getattr(self.hub, "code", None):
            logger.debug("Hub was not changed.")
            return self
        else:
            logger.debug("Hub was changed.")
            return _State(
                self.organizations,
                selected_org_id=self.organization.id,
                selected_hub_code=hub_code,
            )

    def with_workspace(self, workspace_id: UUID | None) -> _State:
        """Create a new state with the specified workspace selected.

        :param workspace_id: The ID of the workspace to select, or `None` to clear the current selection.

        :returns: The new state, or the current state if the workspace was not changed.

        :raises SelectionError: If the workspace is invalid.
        """
        logger.debug("Selecting workspace")
        if workspace_id == getattr(self.workspace, "id", None):
            logger.debug("Workspace was not changed.")
            return self
        else:
            logger.debug("Workspace was changed.")
            return _State(
                self.organizations,
                workspaces=self.workspaces,
                selected_org_id=self.organization.id,
                selected_hub_code=self.hub.code,
                selected_workspace_id=workspace_id,
            )

    def update_organizations(self, organizations: list[Organization]) -> _State:
        """Update the organizations in the state.

        This method will re-select the currently selected organization, hub, and workspace if they are still valid.

        :param organizations: The new list of organizations.

        :returns: The new state with the updated organizations.
        """
        logger.debug("Updating organizations")
        new_state = _State(organizations)
        if self._selected_org_id in new_state._organizations_by_id:
            logger.debug("Re-selecting organization")
            new_state._select_org(self._selected_org_id)

            if self._selected_hub_code in new_state._hubs_by_code:
                logger.debug("Re-selecting hub")
                new_state._select_hub(self._selected_hub_code)

                # Copy the workspaces from the old state to the new state.
                new_state._workspaces_by_id = self._workspaces_by_id.copy()

                if self._selected_workspace_id in new_state._workspaces_by_id:
                    logger.debug("Re-selecting workspace")
                    new_state._select_workspace(self._selected_workspace_id)

        logger.debug("Organizations updated.")
        return new_state

    def update_workspaces(self, workspaces: list[Workspace]) -> _State:
        """Update the workspaces in the state.

        This method will re-select the currently selected workspace if it is still valid.

        :param workspaces: The new list of workspaces.

        :returns: The new state with the updated workspaces.
        """
        logger.debug("Updating workspaces")
        new_state = _State(
            self.organizations,
            workspaces=workspaces,
            selected_org_id=self._selected_org_id,
            selected_hub_code=self._selected_hub_code,
        )

        if self._selected_workspace_id in new_state._workspaces_by_id:
            logger.debug("Re-selecting workspace")
            new_state._select_workspace(self._selected_workspace_id)

        logger.debug("Workspaces updated.")
        return new_state

    def without_workspaces(self) -> _State:
        """Create a new state without any workspaces selected.

        :returns: The new state without any workspaces selected.
        """
        logger.debug("Removing workspaces")
        return _State(
            self.organizations,
            selected_org_id=self._selected_org_id,
            selected_hub_code=self._selected_hub_code,
        )


class ServiceManager:
    """A simple service manager for managing the current selection of organizations, hubs, and workspaces."""

    def __init__(self, transport: ITransport, authorizer: IAuthorizer, discovery_url: str) -> None:
        """
        :param transport: The transport to use for API requests.
        :param authorizer: The authorizer to use for API requests.
        :param discovery_url: The URL of the discovery service.
        """
        self._transport = transport
        self._authorizer = authorizer
        self._discovery_url = discovery_url

        # The state mutex is only required for longer-running operations (e.g., refreshing organizations) to prevent
        # concurrent updates to the state. Read-only operations (e.g., listing organizations) are considered safe
        # because each state instance is not mutated.
        self.__state_mutex = asyncio.Lock()
        self.__state = _State([])

    def _get_connector(self, base_url: str) -> APIConnector:
        return APIConnector(base_url, self._transport, self._authorizer)

    async def refresh_organizations(self) -> None:
        """Refresh the list of organizations.

        This method will update the current state with the latest list of organizations from the discovery service.

        If an error occurs while refreshing the organizations, the state will be reset.
        """
        logger.debug("Refreshing organizations")
        async with self.__state_mutex, self._get_connector(self._discovery_url) as connector:
            client = DiscoveryAPIClient(connector)
            try:
                self.__state = self.__state.update_organizations(await client.list_organizations())
            except Exception:
                logger.exception("Failed to refresh organizations.", exc_info=True)
                self.__state = _State([])
                raise

    def list_organizations(self) -> list[Organization]:
        """Get a list of organizations sorted by display name.

        If organizations have not been refreshed, this method will return an empty list.

        :returns: The list of organizations.
        """
        return self.__state.organizations

    def get_current_organization(self) -> Organization | None:
        """Get the currently selected organization.

        :returns: The currently selected organization, or `None` if no organization is selected.
        """
        return self.__state.organization

    def set_current_organization(self, org_id: UUID | None) -> None:
        """Set the currently selected organization.

        :param org_id: The ID of the organization to select, or `None` to clear the current selection.

        :raises SelectionError: If the organization is invalid.
        """
        self.__state = self.__state.with_organization(org_id)

    def list_hubs(self) -> list[Hub]:
        """Get a list of hubs sorted by display name.

        If no organization is currently selected, this method will return an empty list.

        :returns: The list of hubs.
        """
        return self.__state.hubs

    def get_current_hub(self) -> Hub | None:
        """Get the currently selected hub.

        :returns: The currently selected hub, or `None` if no hub is selected.
        """
        return self.__state.hub

    def set_current_hub(self, hub_code: str) -> None:
        """Set the currently selected hub.

        :param hub_code: The code of the hub to select.

        :raises SelectionError: If the hub is invalid.
        """
        self.__state = self.__state.with_hub(hub_code)

    async def refresh_workspaces(self) -> None:
        """Refresh the list of workspaces.

        This method will update the current state with the latest list of workspaces from the workspace service.

        If an error occurs while refreshing the workspaces, the list of workspaces will be reset.
        """
        logger.debug("Refreshing workspaces")
        async with self.__state_mutex:
            if not isinstance(org := self.get_current_organization(), Organization):
                return  # Cannot refresh workspaces without an organization.

            if not isinstance(hub := self.get_current_hub(), Hub):
                return  # Cannot refresh workspaces without a hub.

            async with self._get_connector(hub.url) as connector:
                client = WorkspaceAPIClient(connector, org.id)
                try:
                    self.__state = self.__state.update_workspaces(await client.list_all_workspaces())
                except Exception:
                    logger.exception("Failed to refresh workspaces.", exc_info=True)
                    self.__state = self.__state.without_workspaces()
                    raise

    def list_workspaces(self) -> list[Workspace]:
        """Get a list of workspaces sorted by display name.

        If workspaces have not been refreshed, this method will return an empty list.

        :returns: The list of workspaces.
        """
        return self.__state.workspaces

    def get_current_workspace(self) -> Workspace | None:
        """Get the currently selected workspace.

        :returns: The currently selected workspace, or `None` if no workspace is selected.
        """
        return self.__state.workspace

    def set_current_workspace(self, workspace_id: UUID) -> None:
        """Set the currently selected workspace.

        :param workspace_id: The ID of the workspace to select.

        :raises SelectionError: If the workspace is invalid.
        """
        self.__state = self.__state.with_workspace(workspace_id)

    def get_connector(self) -> APIConnector:
        """Get an API connector for the currently selected hub.

        :returns: The API connector.

        :raises SelectionError: If no organization or hub is currently selected.
        """
        if not isinstance(self.get_current_organization(), Organization):
            raise SelectionError("No organization is currently selected.")

        if not isinstance(hub := self.get_current_hub(), Hub):
            raise SelectionError("No hub is currently selected.")

        return self._get_connector(hub.url)

    def get_environment(self) -> Environment:
        """Get an environment with the currently selected organization, hub, and workspace.

        :returns: The environment.

        :raises SelectionError: If no organization, hub, or workspace is currently selected.
        """
        if not isinstance(org := self.get_current_organization(), Organization):
            raise SelectionError("No organization is currently selected.")
        if not isinstance(hub := self.get_current_hub(), Hub):
            raise SelectionError("No hub is currently selected.")
        if not isinstance(ws := self.get_current_workspace(), Workspace):
            raise SelectionError("No workspace is currently selected.")
        return Environment(hub_url=hub.url, org_id=org.id, workspace_id=ws.id)

    def create_client(self, client_class: type[T_client], *args: Any, **kwargs: Any) -> T_client:
        """Create a client for the currently selected workspace.

        :param client_class: The class of the client to create.

        :returns: The new client.

        :raises SelectionError: If no organization, hub, or workspace is currently selected.
        """
        return client_class(self.get_environment(), self.get_connector(), *args, **kwargs)



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import asyncio
import contextlib
from collections.abc import AsyncIterator
from typing import Generic, TypeVar

from evo import logging
from evo.common.data import HTTPHeaderDict
from evo.common.interfaces import IAuthorizer

from .connector import OAuthConnector
from .data import AccessToken, AnyScopes, EvoScopes, Scopes
from .exceptions import OAuthError
from .oauth_redirect_handler import OAuthRedirectHandler

__all__ = [
    "AccessTokenAuthorizer",
    "AuthorizationCodeAuthorizer",
    "ClientCredentialsAuthorizer",
]

logger = logging.getLogger("oauth")

T = TypeVar("T", bound=AccessToken)


class _BaseAuthorizer(IAuthorizer, Generic[T]):
    pi_partial_implementation = True  # Suppress warning about missing interface methods.

    def __init__(self, oauth_connector: OAuthConnector, scopes: AnyScopes = EvoScopes.default) -> None:
        """
        :param oauth_connector: The connector to use for fetching tokens.
        :param scopes: The OAuth scopes to request.
        """
        self._mutex = asyncio.Lock()

        self._connector = oauth_connector
        self.__token: T | None = None
        self._scopes = Scopes(scopes)

    def _get_token(self) -> T | None:
        """Get the current access token, or None if the token is not available.

        The mutex should be acquired before setting or getting the token.

        :return: The access token, or None if the token is not available.
        """
        assert self._mutex.locked(), "self._mutex should be acquired before getting the token."
        return self.__token

    def _update_token(self, new_token: T) -> None:
        """Update the access token.

        The mutex should be acquired before setting or getting the token.

        :param new_token: The new access token.
        """
        assert self._mutex.locked(), "self._mutex should be acquired before updating the token."
        self.__token = new_token

    @contextlib.asynccontextmanager
    async def _unwrap_token(self) -> AsyncIterator[T]:
        """A context manager that yields the access token if available.

        :raises OAuthError: If the access token is not available.
        """
        async with self._mutex:
            if self._get_token() is None:
                raise OAuthError("Access token is not available.")
            yield self._get_token()

    async def get_default_headers(self) -> HTTPHeaderDict:
        async with self._unwrap_token() as token:
            return HTTPHeaderDict({"Authorization": f"Bearer {token.access_token}"})


class ClientCredentialsAuthorizer(_BaseAuthorizer[AccessToken]):
    """An OAuth authorizer that uses a client credentials to authorize a client.

    The authorizer will automatically refresh the access token when it expires.
    """

    @contextlib.asynccontextmanager
    async def _unwrap_token(self) -> AsyncIterator[T]:
        """A context manager that yields the access token if available, or fetches a new access token if unavailable.

        :raises OAuthError: If the access token cannot be fetched.
        """
        async with self._mutex:
            if self._get_token() is None:
                self._update_token(await self._fetch_token())
            yield self._get_token()

    async def _fetch_token(self) -> AccessToken:
        """Get an access token from the server.

        https://www.oauth.com/oauth2-servers/access-tokens/client-credentials/

        :return: The access token.

        :raises OAuthError: If token cannot be fetched.
        """
        data = {  # The payload to send to the OAuth server in the token request.
            "grant_type": "client_credentials",
            "scope": self._scopes,
        }

        logger.debug("Fetching access token...")
        return await self._connector.fetch_token(data, AccessToken)

    async def authorize(self) -> None:
        """Get an access token from the server.

        https://www.oauth.com/oauth2-servers/access-tokens/client-credentials/

        :return: The access token.

        :raises OAuthError: If token cannot be fetched.
        """
        async with self._mutex:
            self._update_token(await self._fetch_token())

    async def refresh_token(self) -> bool:
        """Get an access token from the server.

        https://www.oauth.com/oauth2-servers/access-tokens/client-credentials/

        :return: The access token.

        :raises OAuthError: If token cannot be fetched.
        """
        try:
            await self.authorize()
        except Exception:  # noqa
            logger.exception("Failed to refresh the access token.", exc_info=True)
            return False
        else:
            return True


class AuthorizationCodeAuthorizer(_BaseAuthorizer[AccessToken]):
    """An OAuth authorizer that uses a localhost callback to authenticate the user.

    This authorizer launches a local web browser to authenticate the user and obtain an access token. The user is
    required to log in and authorize the application. The authorizer will then use the access token to authorize
    requests.

    The authorizer will automatically refresh the access token when it expires. The refresh token is stored in the
    session and is used to obtain a new access token. If the refresh token expires, the user will need to log in again.
    A mutex is used to ensure that default headers are not returned while the user is logging in or the access token is
    being refreshed.

    This authorizer does not store the access token between sessions. The user will need to log in each time the
    application is started.
    """

    def __init__(
        self, oauth_connector: OAuthConnector, redirect_url: str, scopes: AnyScopes = EvoScopes.default
    ) -> None:
        """
        :param oauth_connector: The OAuth connector to use for fetching tokens.
        :param redirect_url: The local URL to redirect the user back to after authorisation.
        :param scopes: The OAuth scopes to request.
        """
        super().__init__(oauth_connector, scopes)
        self._redirect_url = redirect_url

    def _update_token(self, new_token: AccessToken) -> None:
        return super()._update_token(new_token)

    async def _handle_login(self, timeout_seconds: int) -> AccessToken:
        """Internal method to handle the login process without acquiring the mutex.

        :param timeout_seconds: The maximum time (in seconds) to wait for the authorisation process to complete.

        :return: The new access token.

        :raises OAuthError: If the user does not authenticate within the timeout.
        """
        async with OAuthRedirectHandler(self._connector, self._redirect_url) as handler:
            return await handler.login(scopes=self._scopes, timeout_seconds=timeout_seconds)

    async def login(self, timeout_seconds: int = 180) -> None:
        """Authenticate the user and obtain an access token.

        This method will launch a web browser to authenticate the user and obtain an access token.

        :param timeout_seconds: The maximum time (in seconds) to wait for the authorisation process to complete.

        :raises OAuthError: If the user does not authenticate within the timeout.
        :raises OAuthError: If an error occurred during the authorisation process.
        """
        async with self._mutex:
            new_token = await self._handle_login(timeout_seconds)
            self._update_token(new_token)

    async def refresh_token(self) -> bool:
        async with self._unwrap_token() as old_token, self._connector:
            try:
                if old_token.refresh_token is None:
                    raise OAuthError("Refresh token is missing.")

                data = {  # The payload to send to the OAuth server in the token request.
                    "grant_type": "refresh_token",
                    "refresh_token": old_token.refresh_token,
                }
                logger.debug("Refreshing access token...")
                new_token = await self._connector.fetch_token(data, AccessToken)
                self._update_token(new_token)
            except Exception:  # noqa
                logger.exception("Failed to refresh the access token.", exc_info=True)
                return False
            else:
                return True


class AccessTokenAuthorizer(IAuthorizer):
    def __init__(self, access_token: str):
        """An OAuth authorizer that uses the provided access token to authorize requests.

        This authorizer does not make any attempt to refresh expired tokens. This is handy if you already have an
        access token and are happy to support refreshing it when it expires outside of this interface.
        """
        self._access_token = access_token

    async def refresh_token(self) -> bool:
        logger.debug("AccessTokenAuthorizer does not support refreshing expired tokens.")
        return False

    async def get_default_headers(self) -> HTTPHeaderDict:
        return HTTPHeaderDict({"Authorization": f"Bearer {self._access_token}"})
